open Prolog

let ff = False
let tt = True
let (&&) a b = And (a,b)
let (||) a b = Or (a,b)
let (==) a b = Equals (a,b)

module Test = struct

  let z = Term.make "z" []
  let s x = Term.make "s" [x]

  let leaf = Term.make "leaf" []
  let red n l r = Term.make "red" [n;l;r]
  let black n l r = Term.make "black" [n;l;r]

  let lt n m = Atom ("lt",[n;m])
  let insert x t t' = Atom ("insert",[x;t;t'])
  let balance x l r t = Atom ("balance",[x;l;r;t])
  let is_black t = Atom ("is_black",[t])
  let clean t = Atom ("clean",[t])
  let blackify t t' = Atom ("blackify",[t;t'])
  let make n t = Atom ("make",[n;t])
  let mem x t = Atom ("mem",[x;t])

  let (++) = concat_prog
  let (<==) hd body = horn hd body

  let prog =
    prog_empty
    ++
    make_predicate "lt"
      [ forall (fun x ->
          lt z (s x) <== tt) ;
        forall (fun x -> forall (fun y ->
          lt (s x) (s y) <== lt x y)) ]
    ++
    make_predicate "insert"
      [ (* Insertion à une feuille *)
        forall (fun n ->
          insert n leaf (red n leaf leaf) <== tt) ;
        (* Insertion sous un noeud rouge *)
        forall (fun n -> forall (fun m ->
        forall (fun l -> forall (fun r -> forall (fun l' ->
          insert n (red m l r) (red m l' r) <==
            (lt n m && insert n l l')))))) ;
        forall (fun n -> forall (fun m ->
        forall (fun l -> forall (fun r -> forall (fun r' ->
          insert n (red m l r) (red m l r') <==
            (lt m n && insert n r r')))))) ;
        (* Insertion sous un noeud noir *)
        forall (fun n -> forall (fun m -> forall (fun l -> forall (fun r ->
        forall (fun l' -> forall (fun b ->
          insert n (black m l r) b <==
            (lt n m && insert n l l' && balance m l' r b))))))) ;
        forall (fun n -> forall (fun m -> forall (fun l -> forall (fun r ->
        forall (fun r' -> forall (fun b ->
          insert n (black m l r) b <==
            (lt m n && insert n r r' && balance m l r' b))))))) ]
    ++
    make_predicate "balance"
      [ forall (fun x -> forall (fun y -> forall (fun z ->
        forall (fun a -> forall (fun b ->
        forall (fun c -> forall (fun d ->
          balance z
            (red y (red x a b) c) d
            (red y (black x a b) (black z c d)) <== tt))))))) ;
        forall (fun x -> forall (fun y -> forall (fun z ->
        forall (fun a -> forall (fun b ->
        forall (fun c -> forall (fun d ->
          balance z
            (red x a (red y b c)) d
            (red y (black x a b) (black z c d)) <== tt))))))) ;
        forall (fun x -> forall (fun y -> forall (fun z ->
        forall (fun a -> forall (fun b ->
        forall (fun c -> forall (fun d ->
          balance x
            a (red z (red y b c) d)
            (red y (black x a b) (black z c d)) <== tt))))))) ;
        forall (fun x -> forall (fun y -> forall (fun z ->
        forall (fun a -> forall (fun b ->
        forall (fun c -> forall (fun d ->
          balance x
            a (red y b (red z c d))
            (red y (black x a b) (black z c d)) <== tt))))))) ;
        forall (fun y -> forall (fun l -> forall (fun r ->
          balance y l r (black y l r) <== (clean l && clean r)))) ]
    ++
    make_predicate "clean"
      [ forall (fun t -> clean t <== is_black t) ;
        forall (fun x -> forall (fun l -> forall (fun r ->
          clean (red x l r) <== (is_black l && is_black r)))) ]
    ++
    make_predicate "is_black"
      [ is_black leaf <== tt ;
        forall (fun x -> forall (fun y -> forall (fun z ->
          is_black (black x y z) <== tt))) ]
    ++
    make_predicate "blackify"
      [ forall (fun x -> forall (fun l -> forall (fun r ->
          blackify (black x l r) (black x l r) <== tt))) ;
        forall (fun x -> forall (fun l -> forall (fun r ->
          blackify (red x l r) (black x l r) <== tt))) ]
    ++
    (* Recherche *)
    make_predicate "mem"
      [ forall (fun x -> forall (fun l -> forall (fun r ->
          mem x (black x l r) <== tt))) ;
        forall (fun x -> forall (fun l -> forall (fun r ->
          mem x (red x l r) <== tt))) ;
        forall (fun x -> forall (fun y -> forall (fun l -> forall (fun r ->
          mem x (black y l r) <== ((lt x y && mem x l) ||
                                   (lt y x && mem x r)))))) ;
        forall (fun x -> forall (fun y -> forall (fun l -> forall (fun r ->
          mem x (red y l r) <== ((lt x y && mem x l)
                              || (lt y x && mem x r)))))) ]
    ++
    (* Insertion des entiers de 1 à n. *)
    make_predicate "make"
      [ forall (fun t ->
          make z leaf <== tt) ;
        forall (fun n -> forall (fun t ->
        forall (fun t' -> forall (fun t'' ->
          make (s n) t'' <==
            (make n t && insert (s n) t t' && blackify t' t''))))) ]


  (** Exécution des tests sur l'interprète donné en argument. *)
  let run has_solution =

    Gc.set { (Gc.get ()) with
               Gc.minor_heap_size = 2048000 ;
               Gc.space_overhead = max_int } ;

    let has_solution = has_solution prog in

    let rec of_nat = function
      | 0 -> z
      | n -> s (of_nat (n-1))
    in

    let x = Term.var (Term.fresh ()) in
    let t0 = Unix.gettimeofday () in
    assert (has_solution (make (of_nat 0) x)) ;
    assert (has_solution (make (of_nat 1) x)) ;
    assert (has_solution (make (of_nat 5) x)) ;
    assert (has_solution (make (of_nat 10) x && mem (of_nat 6) x)) ;
    assert (not (has_solution (make (of_nat 10) x && mem (of_nat 0) x))) ;
    assert (not (has_solution (make (of_nat 10) x && mem (of_nat 11) x))) ;
    let n = 500 in
    assert (not (has_solution (make (of_nat n) x && mem (of_nat 0) x))) ;
    Format.printf "[OK] Tests passed (%.2fs)\n" (Unix.gettimeofday () -. t0)

end

let () =
  Format.printf "Testing abstract machine...\n%!" ;
  Test.run Abstract_machine.has_solution
