
(** Les buts sont des formules logiques simples.
  * Le sens des atomes sera donné juste après par un programme logique. *)

type args = Term.in_term list

type goal =
  | Atom of string * args
  | Equals of Term.in_term * Term.in_term
  | And of goal * goal
  | Or of goal * goal
  | False
  | True

(** Programmes logiques.
  * On a un type abstrait et une seule opération permettant
  * d'extraire le but associé à un atome. *)

type program
val lookup : program -> string -> args -> goal

(** Constructions élémentaires de programmes.
  * La concaténation doit fonctionner quand les deux programmes
  * définissent des prédicats disjoints. Si ce n'est pas le cas
  * son comportement n'est pas spécifié, mais le mieux est de lever
  * une erreur. *)

val prog_empty : program
val concat_prog : program -> program -> program

(** Construction du programme correspondant à un prédicat,
  * à partir d'un ensemble de clauses, elles-mêmes obtenues
  * par un certain nombre de quantifications universelles
  * (forall) sur une implication (horn). *)

type clause
val horn : goal -> goal -> clause
val forall : (Term.in_term -> clause) -> clause

val make_predicate : string -> clause list -> program
