(** Le type in_term correspond à la représentation interne des termes.
  * Le type var représente les variables, c'est à dire les objets que
  * l'on peut instantier.
  * Le type term correspond à un terme superficiellement explicité. *)

type in_term
type var
type term = Fun of string * in_term list | Var of var

(** Modification d'une variable. *)
val bind : var -> in_term -> unit

(** Observation d'un terme. *)
val observe : in_term -> term

(** Egalité syntaxique entre termes et variables. *)

val equals : in_term -> in_term -> bool
val var_equals : var -> var -> bool


(** Constructeurs de termes. *)

(** Création d'un terme construit à partir d'un symbole
  * de fonction -- ou d'une constante, cas d'arité 0. *)
val make : string -> in_term list -> in_term

(** Création d'un terme restreint à une variable. *)
val var : var -> in_term

(** Création d'une variable fraîche. *)
val fresh : unit -> var


(** Manipulation de l'état: sauvegarde, restauration. *)

type state

(** [save ()] renvoie un descripteur de l'état actuel. *)
val save : unit -> state

(** [restore s] restaure les variables dans l'état décrit par [s]. *)
val restore : state -> unit
